from typing import Any, Optional
from pprint import pformat

from django.core.exceptions import ValidationError
from django.utils.safestring import mark_safe

from .validators import call_validator
from . import PREVIEW_HTML, pre, PREVIEW_TEXT, PREVIEW_PYTHON, PREVIEW_NONE


def mix(*cls):
    return type("", cls, {})


class MinMaxValidationMixin:
    min_value = None
    max_value = None

    def validate(self, value):
        super().validate(value)

        is_none_valid = self.min_value is None and self.max_value is None
        if not is_none_valid and value is None:
            raise ValidationError("Value cannot be None")

        if self.min_value is not None and value < self.min_value:
            raise ValidationError(f"Value cannot be less than {self.min_value}")

        if self.max_value is not None and value > self.max_value:
            raise ValidationError(f"Value cannot be greater than {self.max_value}")

    def get_help_format(self):
        yield from super().get_help_format()

        if self.min_value is not None:
            yield f" from {self.min_value}"

        if self.max_value is not None:
            yield f" to {self.max_value}"


class EmptyNoneMixin:
    value_required = False

    def to_python(self, value):
        if value.strip() == "":
            return None
        return super().to_python(value)


class HTMLMixin:
    admin_preview_as: str = PREVIEW_HTML

    def get_help_format(self):
        yield from super().get_help_format()
        yield " in HTML format"

    def give(self, value, suffix=None):
        return mark_safe(super().give(value, suffix))


class PositiveValidationMixin(MinMaxValidationMixin):
    min_value = 0


class CallToPythonMixin:
    call_func = None

    def prepare_python_call(self, value):
        return {"prepared": value}

    def get_preview_validators(self):
        return [
            validator
            for validator in self.get_validators()
            if isinstance(validator, call_validator)
        ]

    def get_call_func(self):
        return self.call_func

    def python_call(self, *args, **kwargs):
        return self.call_func(*args, **kwargs)

    def to_python(self, value):
        value = super().to_python(value)
        prepared_kwargs = self.prepare_python_call(value)

        def _(*args, **kwargs):
            return self.python_call(*args, **prepared_kwargs, **kwargs)

        return _

    def give_python_to_admin(self, value, name, **kwargs):
        try:
            value = self.to_python(value)
        except Exception as e:
            return [(None, e)]
        ret = []
        for validator in self.get_preview_validators():
            str_validator = f"{name}({str(validator)})"
            try:
                ret.append((str_validator, validator(value)))
            except Exception as e:
                ret.append((str_validator, e))
        return ret

    def get_admin_preview_object(self, value, name, **kwargs):
        admin_preview_as = self.get_admin_preview_as()
        if admin_preview_as == PREVIEW_NONE:
            return ""

        if not value:
            return "No preview (add at least one call_validator in validators)"

        if len(value) == 1 and admin_preview_as in (PREVIEW_TEXT, PREVIEW_HTML):
            value = value[0][1]
            if isinstance(value, Exception):
                return f"ERROR!!! {value}"
            if admin_preview_as == PREVIEW_TEXT:
                return pre(value)
            else:
                return str(value)

        def _():
            for validator, val in value:
                yield pre(f">>> {validator}")

                if isinstance(val, Exception):
                    yield f"ERROR!!! {val}"
                else:
                    if admin_preview_as == PREVIEW_TEXT:
                        yield pre(val)
                    elif admin_preview_as == PREVIEW_HTML:
                        yield str(val)
                    else:
                        yield pre("<<< " + pformat(val))

        return "\n".join(_())


class GiveCallMixin:
    """
    for Template-like classes, when to_python should return a function,
    but as value you want to return not a function, but call it and return result
    """

    def get_suffixes(self):
        return ("call",) + super().get_suffixes()

    def get_validators(self):
        return (call_validator(),) + super().get_validators()

    def get_validators(self):
        validators = super().get_validators()
        if not validators:
            return (call_validator(),)
        return validators

    def give(self, value, suffix=None):
        if suffix is None:
            return value()
        if suffix == "call":
            return value
        return super().give(value, suffix)


class MakeCallMixin:
    def get_suffixes(self):
        return ("call",) + super().get_suffixes()

    def give_python_to_admin(self, value, name, **kwargs):
        return self.give_python(value)()

    def give(self, value, suffix=None):
        return lambda *args, **kwargs: value


class DictSuffixesMixin:
    suffixes = {}

    def give(self, value, suffix=None):
        if suffix is None:
            return value
        return self.suffixes[suffix](value)


class AdminPreviewMenuMixin:
    def gen_admin_preview_html_menu_items(self, value: Any, name: str, **kwargs) -> str:
        return
        yield

    def get_admin_preview_html_menu(self, value: Any, name: str, **kwargs) -> str:
        html_items = "".join(
            self.gen_admin_preview_html_menu_items(value, name, **kwargs)
        )
        if not html_items:
            return ""
        return "<div>" + html_items + "</div>"

    def get_admin_preview_object(self, value: str, name: str, **kwargs) -> str:

        return self.get_admin_preview_html_menu(
            value, name, **kwargs
        ) + self.get_admin_preview_under_menu_object(value, name, **kwargs)

    def get_admin_preview_under_menu_object(
        self, value: Any, name: str, **kwargs
    ) -> str:
        return super().get_admin_preview_object(value, name, **kwargs)


class AdminSuffixesMixinPreview:
    admin_preview_suffixes = ()
    admin_preview_suffixes_default = "default"

    def get_admin_preview_suffixes_default(self):
        return self.admin_preview_suffixes_default

    def get_admin_preview_suffixes(self, value: str, name: str, **kwargs):
        return self.admin_preview_suffixes

    def gen_admin_preview_html_menu_items(
        self, value: Any, name: str, suffix: Optional[str] = None, **kwargs
    ) -> str:
        suffixes = self.get_admin_preview_suffixes(value, name, **kwargs)
        if not len(suffixes):
            return

        for suf in (None, *suffixes):
            if suf == suffix:
                yield f" <b>{suf or self.get_admin_preview_suffixes_default()}</b> "
            elif suf is None:
                yield f' <a class="cs_set_params">{self.get_admin_preview_suffixes_default()}</a> '
            else:
                yield f' <a class="cs_set_params" data-param-suffix="{suf}">{suf}</a> '

        yield from super().gen_admin_preview_html_menu_items(
            value, name, suffix=suffix, **kwargs
        )

    def get_admin_preview_under_menu_object(
        self, value: Any, name: str, **kwargs
    ) -> str:
        suffix = kwargs.get("suffix", None)
        return super().get_admin_preview_under_menu_object(
            self.give(value, suffix), name, **kwargs
        )


class AdminPreviewSuffixesMixin(AdminSuffixesMixinPreview, AdminPreviewMenuMixin):
    pass


class ActionResponse(dict):
    def error(self, message):
        self["error"] = message
        return self

    def html(self, html):
        self["html"] = html
        return self

    def alert(self, message):
        self["alert"] = message
        return self

    def value(self, value):
        self["value"] = value
        return self

    def before_html(self, html):
        self["before_html"] = html
        return self


class AdminActionsMixinPreview:
    # list of tuples (name, function)
    admin_preview_actions = None

    def gen_admin_preview_html_menu_items(
        self, value: Any, name: str, suffix: Optional[str] = None, **kwargs
    ) -> str:
        if self.admin_preview_actions is None:
            return

        for i, (action_name, _) in enumerate(self.admin_preview_actions):
            yield f' <a class="cs_set_params cs_set_params_once" data-param-action="{i}">{action_name}</a> '

    def get_admin_preview_function(self, *args, **kwargs):
        if self.admin_preview_actions is None:
            return

        action_name = kwargs.get("action", None)
        if action_name is None:
            return

        try:
            return self.admin_preview_actions[int(action_name)][1]
        except (IndexError, ValueError):
            return

    def process_action(self, *args, **kwargs):
        func = self.get_admin_preview_function(*args, **kwargs)
        if func is None:
            return

        response = ActionResponse()
        func(response, *args, _cs_type=self, **kwargs)
        return response

    def get_full_admin_preview_value(self, *args, **kwargs) -> str:
        response = self.process_action(*args, **kwargs)
        if not response:
            return super().get_full_admin_preview_value(*args, **kwargs)

        response = dict(response)

        before_html = response.pop("before_html", "")
        if before_html:
            response = {
                **response,
                **super().get_full_admin_preview_value(*args, **kwargs),
            }
            response["html"] = before_html + response["html"]

        return response


class AdminPreviewActionsMixin(AdminActionsMixinPreview, AdminPreviewMenuMixin):
    pass


class DictSuffixesPreviewMixin(DictSuffixesMixin, AdminPreviewSuffixesMixin):
    def get_admin_preview_suffixes(self, value: str, name: str, **kwargs):
        return tuple(self.suffixes.keys())



PREVIEW_HTML = "html"
PREVIEW_TEXT = "text"
PREVIEW_PYTHON = "python"
PREVIEW_NONE = "none"

PREVIEW_ALL = [
    value for name, value in globals().items() if name.startswith("PREVIEW_")
]


class required:
    pass


class optional:
    pass


def pre(value: str) -> str:
    return "<pre>{}</pre>".format(str(value).replace("<", "&lt;"))



from typing import Any, Optional

from django.core.exceptions import ValidationError

from .basic import (
    SimpleText,
    SimpleTextPreview,
    SimpleString,
    BaseSetting,
    PREVIEW_PYTHON,
    PREVIEW_TEXT,
    PREVIEW_NONE,
    PREVIEW_HTML,
)
from .mixins import AdminPreviewSuffixesMixin
from .each import EachMixin, Item, Keys, Values


def f_empty(value):
    value = value.strip()
    if not value:
        return None
    return value


def f_comment(prefix):
    def _(value):
        if value.strip().startswith(prefix):
            return None
        return value

    return _


class SimpleStringsList(SimpleText):
    comment_starts_with = "#"
    filter_empty = True
    split_lines = "\n"
    filters = None
    admin_preview_as = PREVIEW_PYTHON

    def get_filters(self):
        if self.filters is not None:
            return self.filters

        filters = []
        if self.filter_empty:
            filters.append(f_empty)

        if self.comment_starts_with:
            filters.append(f_comment(self.comment_starts_with))

        return filters

    def get_help_format(self):
        yield "List of values with the following format:"
        yield "<ul>"
        if self.split_lines == "\n":
            yield "<li>each line is a new value</li>"
        else:
            yield f"<li>{self.split_lines} separates values</li>"
        yield "<li>strip spaces from the beginning and from the end of the value</li>"
        yield "<li>remove empty values</li>"
        if self.comment_starts_with:
            yield f"<li> use {self.comment_starts_with} to comment a line</li>"
        if self.filter_empty:
            yield "<li>empty values are removed</li>"
        yield "</ul>"

    def filter_line(self, line):
        for f in self.get_filters():
            line = f(line)
            if line is None:
                break
        return line

    def gen_to_python(self, value):
        lines = value.split(self.split_lines)
        for line in lines:
            line = self.filter_line(line)

            if line is not None:
                yield line

    def to_python(self, value):
        return list(self.gen_to_python(value))


class TypedStringsList(EachMixin, SimpleStringsList):
    line_type = SimpleString()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.each = Item(self.line_type)


NOT_FOUND_DEFAULT = "default"
NOT_FOUND_KEY_ERROR = "key_error"
NOT_FOUND_VALUE = "value"

SPLIT_FAIL_IGNORE = "ignore"
SPLIT_FAIL_RAISE = "raise"


def split_validator_in(values):
    def _(value):
        return value in values

    return _


class SplitTextByFirstLine(SimpleText):
    split_default_key = None
    split_default_chooser = None
    split_not_found = NOT_FOUND_DEFAULT
    split_not_found_value = None
    split_key_validator = None
    split_key_validator_failed = SPLIT_FAIL_IGNORE
    admin_preview_as = PREVIEW_HTML

    def get_split_default_key(self):
        return self.split_default_key

    def split_default_choose(self, value):
        if self.split_default_chooser is None:
            return self.get_split_default_key()
        else:
            return self.split_default_chooser(value)

    def get_split_not_found(self):
        return self.split_not_found

    def get_split_not_found_value(self):
        return self.split_not_found_value

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert (
            self.get_split_default_key()
        ), "split_default_key should be set and it should not be empty"
        assert isinstance(
            self.get_split_default_key(), str
        ), "split_default_key should be str"

        assert self.split_default_chooser is None or callable(
            self.split_default_chooser
        ), "split_default_chooser should be callable or None"
        assert self.split_key_validator is None or callable(
            self.split_key_validator
        ), "split_key_validator should be callable or None"
        assert self.split_key_validator_failed in (
            SPLIT_FAIL_IGNORE,
            SPLIT_FAIL_RAISE,
        ), "split_key_validator_failed should be one of SPLIT_FAIL_IGNORE, SPLIT_FAIL_RAISE"
        assert self.get_split_not_found() in (
            NOT_FOUND_DEFAULT,
            NOT_FOUND_KEY_ERROR,
            NOT_FOUND_VALUE,
        ), "split_not_found should be one of NOT_FOUND_DEFAULT, NOT_FOUND_KEY_ERROR, NOT_FOUND_VALUE"

    def validate_split_key(self, key):
        return self.split_key_validator is None or self.split_key_validator(key)

    def split_value(self, value):
        lines = value.splitlines()
        if (
            not lines
            or self.get_split_default_key() not in lines[0]
            or lines[0].count(self.get_split_default_key()) > 1
        ):
            return {self.get_split_default_key(): value}

        before, after = lines[0].split(self.get_split_default_key())
        ret = {self.get_split_default_key(): []}
        cur_iter = ret[self.get_split_default_key()]
        for num, line in enumerate(lines[1:], start=2):
            if line.startswith(before) and line.endswith(after):
                line_key = line[len(before) : len(line) - len(after)]
                if self.validate_split_key(line_key):
                    cur_iter = ret[line_key] = []
                    continue

                if self.split_key_validator_failed == SPLIT_FAIL_RAISE:
                    raise ValidationError(
                        f"Invalid split key: {line_key} in line {num}"
                    )

            cur_iter.append(line)

        return {k: "\n".join(v) for k, v in ret.items()}

    def to_python(self, value):
        return self.split_value(value)

    def give(self, value, suffix=None):
        if suffix is None:
            suffix = self.split_default_choose(value)

        try:
            ret_val = value[suffix.upper()]
        except KeyError:
            if self.get_split_not_found() == NOT_FOUND_KEY_ERROR:
                raise
            elif self.get_split_not_found() == NOT_FOUND_DEFAULT:
                ret_val = value[self.get_split_default_key()]
            elif self.get_split_not_found() == NOT_FOUND_VALUE:
                ret_val = self.get_split_not_found_value()

        return ret_val


class SplitByFirstLine(AdminPreviewSuffixesMixin, EachMixin, SplitTextByFirstLine):
    split_type = SimpleTextPreview()

    def get_admin_preview_suffixes_default(self):
        return self.get_split_default_key()

    def get_admin_preview_suffixes(self, value: dict, name: str, **kwargs):
        keys = list(value.keys())
        keys.remove(self.get_split_default_key())
        return tuple(keys)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if isinstance(self.split_type, dict):
            self.each = Keys(**self.split_type)
        else:
            self.each = Values(self.split_type)

    def get_admin_preview_under_menu_object(self, value, name, suffix=None, **kwargs):
        if suffix is None:
            suffix = self.get_split_default_key()
        cs_type = self.split_type
        if isinstance(cs_type, dict):
            cs_type = self.split_type.get(suffix.upper(), SimpleTextPreview())

        value = SplitTextByFirstLine.give(self, value, suffix)
        return cs_type.get_admin_preview_object(value, name, **kwargs)


class SplitTranslation(SplitByFirstLine):
    split_default_key = "EN"
    split_not_found = NOT_FOUND_DEFAULT

    def get_help_format(self):
        yield "Translated Text. "
        yield "The first line can initialize the translation splitter. "
        yield f"The initial language is {self.split_default_key}. "
        yield f"The first line can be '===== {self.split_default_key} ====='. "
        yield "The format for the value inside the translation is: "
        yield from self.split_type.get_help_format()

    def split_default_choose(self, value):
        from django.utils.translation import get_language

        return get_language().upper().replace("-", "_")



from inspect import isclass
from functools import cached_property
from pprint import pformat
from typing import Optional, Set, Tuple, Union, Any, Callable
from collections.abc import Iterable
from json import dumps

from django import forms
from django.core.serializers.json import DjangoJSONEncoder
from django.core.exceptions import ValidationError

from content_settings.context_managers import context_defaults_kwargs
from content_settings.settings import CACHE_SPLITER
from content_settings.types.lazy import LazyObject
from content_settings.types import (
    PREVIEW_ALL,
    PREVIEW_HTML,
    PREVIEW_TEXT,
    PREVIEW_PYTHON,
    PREVIEW_NONE,
    required,
    optional,
    pre,
)
from content_settings.types.mixins import HTMLMixin
from content_settings.permissions import none, staff


class BaseSetting:
    pass


class SimpleString(BaseSetting):
    """
     Attributes:
    - constant (bool): Whether the setting is constant (can not be changed).
    - cls_field (forms.CharField): The form field class to use for the setting.
    - widget (forms.Widget): The form widget to use for the cls_field.
    - widget_attrs (Optional[dict]): Optional attributes for the widget initiation.
    - fetch_permission (Callable): Optional permission required to fetch the setting.
    - update_permission (Callable): Optional permission required to update the setting.
    - help_format (Optional[str]): Optional format string for the help text for the format (align to the type).
    - help (Optional[str]): Optional help text for the setting.
    - value_required (bool): Whether a value is required for the setting.
    - version (str): The version of the setting (using for caching).
    - tags (Optional[Iterable[str]]): Optional tags associated with the setting.
    - validators (Tuple[Callable]): Validators to apply to the setting value.
    - validators_raw (Tuple[Callable]): Validators to apply to the text value of the setting.
    - admin_preview_as (str): The format to use for the admin preview.
    - suffixes (Tuple[str]): Suffixes that can be appended to the setting value.
    - user_defined_slug (str): it contains a slug from db If the setting is defined in DB only (should not be set in content_settings)
    - overwrite_user_defined (bool): Whether the setting can overwrite a user defined setting.
    - default (str): The default value for the setting.
    """

    constant: bool = False
    cls_field: forms.CharField = forms.CharField
    widget: forms.Widget = forms.TextInput
    widget_attrs: Optional[dict] = None
    fetch_permission: Callable = staticmethod(none)
    update_permission: Callable = staticmethod(staff)
    view_permission: Callable = staticmethod(staff)
    view_history_permission: Optional[Callable] = None
    help_format: str = "string"
    help: str = ""
    value_required: bool = False
    version: str = ""
    tags: Optional[Iterable[str]] = None
    validators: Tuple[Callable] = ()
    validators_raw: Tuple[Callable] = ()
    admin_preview_as: str = PREVIEW_NONE
    suffixes: Tuple[str] = ()
    user_defined_slug: Optional[str] = None
    overwrite_user_defined: bool = False
    default: Union[str, required, optional] = ""
    json_encoder = DjangoJSONEncoder

    def __init__(
        self, default: Optional[Union[str, required, optional]] = None, **kwargs
    ):
        # optional support of help_text instead of help
        if "help" not in kwargs and "help_text" in kwargs:
            kwargs["help"] = kwargs.pop("help_text")

        for k in kwargs.keys():
            assert self.can_assign(k), "Attribute {} not found".format(k)

        if default is not None:
            self.default = default
        assert self.default in (required, optional) or isinstance(
            self.default, str
        ), "Default should be str (or required or optional for special cases)"

        kwargs = self.update_defaults_context(kwargs)
        self.init_assign_kwargs(kwargs)

        assert isinstance(self.version, str), "Version should be str"
        assert (
            CACHE_SPLITER not in self.version
        ), f"Version should not contain CACHE_SPLITER:{CACHE_SPLITER}"
        assert (
            self.get_admin_preview_as() in PREVIEW_ALL
        ), f"admin_preview_as should be in {PREVIEW_ALL}"

    def can_view(self, user):
        return self.view_permission(user)

    def can_view_history(self, user):
        return (
            self.can_view(user)
            if self.view_history_permission is None
            else self.view_history_permission(user)
        )

    def can_update(self, user):
        return self.update_permission(user)

    def can_fetch(self, user):
        return self.fetch_permission(user)

    def get_admin_preview_as(self) -> str:
        return self.admin_preview_as

    def init_assign_kwargs(self, kwargs):
        for k, v in kwargs.items():
            if not self.can_assign(k):
                raise ValueError("Attribute {} not found".format(k))
            if hasattr(self, "init__{}".format(k)):
                getattr(self, "init__{}".format(k))(v)
            else:
                setattr(self, k, v)

    def init__tags(self, tags: Union[None, str, Iterable[str]]) -> None:
        if not tags:
            return

        if isinstance(tags, str):
            tags = {tags}
        else:
            tags = set(tags)

        self.tags = tags if self.tags is None else self.tags | tags

    @cached_property
    def field(self) -> forms.Field:
        return self.get_field()

    def get_suffixes(self) -> Tuple:
        return self.suffixes

    def can_suffix(self, suffix: Optional[str]) -> bool:
        return suffix is None or suffix in self.get_suffixes()

    def can_assign(self, name: str) -> bool:
        return hasattr(self, name)

    def update_defaults_context(self, kwargs: dict) -> dict:
        # initiate default values
        context_defaults = {
            name: value
            for name, value in context_defaults_kwargs().items()
            if self.can_assign(name)
        }

        # update actual values
        return {
            name: value
            for name, value in context_defaults_kwargs(
                {
                    **context_defaults,
                    **kwargs,
                }
            ).items()
            if self.can_assign(name)
        }

    def get_help_format(self) -> Iterable[str]:
        yield self.help_format

    def get_help(self) -> str:
        help = self.help
        help_format = "".join(self.get_help_format())

        if help and help_format:
            return f"{help}<br><br>{help_format}"

        return help or help_format

    def get_tags(self) -> Set[str]:
        tags = self.tags
        if tags is None:
            tags = set()
        return set(tags)

    def get_content_tags(self, value: str) -> Set[str]:
        from content_settings.conf import gen_tags

        return gen_tags(self, value)

    def get_validators(self) -> Tuple[Callable]:
        return tuple(self.validators) + tuple(self.cls_field.default_validators)

    def get_validators_raw(self) -> Tuple[Callable]:
        return tuple(self.validators_raw)

    def get_field(self) -> forms.Field:
        return self.cls_field(
            widget=self.get_widget(),
            validators=(self.validate_value,),
            required=self.value_required,
        )

    def get_widget(self) -> forms.Widget:
        if isclass(self.widget) and self.widget_attrs is not None:
            return self.widget(attrs=self.widget_attrs)
        else:
            return self.widget

    def validate_raw_value(self, value: str) -> None:
        for validator in self.get_validators_raw():
            validator(value)

    def validate_value(self, value: str) -> Any:
        if self.value_required and not value:
            raise ValidationError("This field is required")
        self.validate_raw_value(value)
        val = self.to_python(value)
        self.validate(val)

    def validate(self, value):
        for validator in self.get_validators():
            validator(value)

    def to_python(self, value: str) -> Any:
        return self.field.to_python(value)

    def json_view_value(self, value: Any, **kwargs) -> Any:
        return dumps(value, cls=self.json_encoder)

    def give_python_to_admin(self, value: str, name: str, **kwargs) -> Any:
        return self.to_python(value)

    def get_admin_preview_html(self, value: Any, name: str, **kwargs) -> Any:
        return value

    def get_admin_preview_text(self, value: Any, name: str, **kwargs) -> Any:
        return value

    def get_admin_preview_python(self, value: Any, name: str, **kwargs) -> Any:
        return value

    def get_admin_preview_value(self, value: str, *args, **kwargs) -> str:
        if self.get_admin_preview_as() == PREVIEW_NONE:
            return ""

        value = self.give_python_to_admin(value, *args, **kwargs)

        return str(self.get_admin_preview_object(value, *args, **kwargs))

    def get_full_admin_preview_value(self, value: str, *args, **kwargs) -> str:
        try:
            self.validate_value(value)
        except Exception as e:
            return {
                "error": str(e),
            }

        return {
            "html": self.get_admin_preview_value(value, *args, **kwargs),
        }

    def get_admin_preview_object(self, value: str, name: str, **kwargs) -> str:
        if self.get_admin_preview_as() == PREVIEW_HTML:
            return str(self.get_admin_preview_html(value, name, **kwargs))

        if self.get_admin_preview_as() == PREVIEW_TEXT:
            value = str(self.get_admin_preview_text(value, name, **kwargs))
        else:
            value = pformat(self.get_admin_preview_python(value, name, **kwargs))

        return pre(value)

    def lazy_give(self, l_func: Callable, suffix=None) -> LazyObject:
        return LazyObject(l_func)

    def give(self, value, suffix: Optional[str] = None):
        return value

    def give_python(self, value: str, suffix=None) -> Any:
        return self.give(self.to_python(value), suffix)


class SimpleText(SimpleString):
    widget: forms.Widget = forms.Textarea
    widget_attrs: dict = {"rows": 10, "cols": 80}


class SimpleTextPreview(SimpleText):
    admin_preview_as: str = PREVIEW_TEXT


class SimpleHTML(HTMLMixin, SimpleText):
    pass


class URLString(SimpleString):
    cls_field: forms.Field = forms.URLField
    widget: forms.Widget = forms.URLInput
    help_format: str = "URL"
    widget_attrs: dict = {"style": "max-width: 600px; width: 100%"}


class EmailString(SimpleString):
    cls_field: forms.Field = forms.EmailField
    widget: forms.Widget = forms.EmailInput
    help_format: str = "Email"
    widget_attrs: dict = {"style": "max-width: 600px; width: 100%"}


class SimpleInt(SimpleString):
    admin_preview_as: str = PREVIEW_PYTHON
    cls_field: forms.Field = forms.IntegerField
    help_format: str = "Any number"


class SimpleBool(SimpleString):
    admin_preview_as: str = PREVIEW_PYTHON
    yeses: Tuple[str] = ("yes", "true", "1", "+", "ok")
    noes: Tuple[str] = ("no", "not", "false", "0", "-", "")

    def to_python(self, value) -> bool:
        value = value.lower().strip()
        if value in self.yeses:
            return True
        if value in self.noes:
            return False
        return None

    @property
    def explained_accepted_values(self):
        return [f"'{v}'" if v else "empty" for v in self.yeses + self.noes]

    def validate(self, value):
        if value is None:
            raise ValidationError(
                f"Value cannot be {', '.join(self.explained_accepted_values)} only"
            )

    def get_help_format(self):
        yield "boolean (True/False) value. "
        yield f"Accepted values: {', '.join(self.explained_accepted_values)}"


class SimpleDecimal(SimpleString):
    admin_preview_as: str = PREVIEW_PYTHON
    cls_field: forms.Field = forms.DecimalField
    help_format: str = "Decimal number with floating point"


class SimplePassword(SimpleString):
    admin_preview_as: str = PREVIEW_NONE
    widget_attrs: dict = {"type": "password"}
